/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v1

// PodSchedulingGroupApplyConfiguration represents a declarative configuration of the PodSchedulingGroup type for use
// with apply.
//
// PodSchedulingGroup identifies the runtime scheduling group instance that a Pod belongs to.
// The scheduler uses this information to apply workload-aware scheduling semantics.
// Exactly one field must be specified.
type PodSchedulingGroupApplyConfiguration struct {
	// PodGroupName specifies the name of the standalone PodGroup object
	// that represents the runtime instance of this group.
	// Must be a DNS subdomain.
	PodGroupName *string `json:"podGroupName,omitempty"`
}

// PodSchedulingGroupApplyConfiguration constructs a declarative configuration of the PodSchedulingGroup type for use with
// apply.
func PodSchedulingGroup() *PodSchedulingGroupApplyConfiguration {
	return &PodSchedulingGroupApplyConfiguration{}
}

// WithPodGroupName sets the PodGroupName field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the PodGroupName field is set to the value of the last call.
func (b *PodSchedulingGroupApplyConfiguration) WithPodGroupName(value string) *PodSchedulingGroupApplyConfiguration {
	b.PodGroupName = &value
	return b
}
